// sfc32 implementation in lua, based on PractRand's sfc.cpp
// src/RNGs/sfc.cpp

if kartrand then return end // already loaded

print("\x8C".."KartRand: ".."\x80".."based on ".."\x88".."sfc32!")

// simple helper for l8er
rawset(_G, "deepcopy", function(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end)

// Various magic numbers
// good sets include {21,9,3},{15,8,3}; older versions used {25,8,3} which wasn't as good
local BARREL_SHIFT = 21
local RSHIFT = 9
local LSHIFT = 3

local function internal_random(state)
	// increment our counter
	state.d = $ + 1
	state.e = state.a + state.b + state.d
	state.a = state.b ^^ (state.b >> RSHIFT)
	state.b = state.c + (state.c << LSHIFT)
	state.c = (($ << BARREL_SHIFT) | ($ >> (32-BARREL_SHIFT))) + state.e
	return state.e
end

// capped version of above
local function internal_capped(s, limit)
	if limit < 1 then
		print("limit was "..limit.." but must be 1 or larger")
		return 0
	end
	local maximum_gen = (INT32_MAX / limit)*limit
	local value
	repeat
		value = internal_random(s)
	until (value >= 0) and (value <= maximum_gen)
	return value % limit
end

rawset(_G, "KR_RandomState", function(s)
	// d is counter in the original cpp file
	local finalstate = {a = 0, b = 0, c = 0, d = 0, e = 0}
	finalstate.b = s >> 0
	finalstate.c = s >> 32
	finalstate.d = 1
	// repeat for a certain amount of times
	for i=0,8,1 do
		internal_random(finalstate)
	end
	return finalstate
end)

// init our global state
local kartrand_global = KR_RandomState(6969420)

// REIMPLEMENT EVERY SINGLE POSSIBLE RANDOM FUNCTION

// P_Random()
rawset(_G, "KR_Random", function(s)
	s = $ or kartrand_global
	return internal_random(s)
end)

// P_RandomFixed()
rawset(_G, "KR_RandomFixed", function(s)
	s = $ or kartrand_global
	return internal_random(s) & 65535
end)

// P_RandomByte()
rawset(_G, "KR_RandomByte", function(s)
	s = $ or kartrand_global
	return internal_random(s) & 255
end)

// P_RandomByte() but with a range of [-128, 128]
rawset(_G, "KR_RandomSignedByte", function(s)
	s = $ or kartrand_global
	return (internal_random(s) & 255)-128
end)

// idk
rawset(_G, "KR_RandomBool", function(s)
	s = $ or kartrand_global
	return (internal_random(s) & 1) == 1
end)

local function getstatefromvarargs(tbl)
	// dunno if this even works lmao
	local n_tbl = deepcopy(tbl)
	for k, v in ipairs(tbl) do
		if type(v) == "table" then
			local val = table.remove(n_tbl, k)
			return val, n_tbl 
		end
	end
	return kartrand_global, n_tbl
end

// P_RandomChance()
rawset(_G, "KR_RandomChance", function(...)
	local s, args = getstatefromvarargs({...})
	local chance_frac = args[1]
	if chance_frac >= FRACUNIT then return true
	elseif chance_frac <= 0 then return false end
	return (internal_random(s) & 65535) >= chance_frac
end)

// P_RandomKey()
rawset(_G, "KR_RandomKey", function(...)
	local s, args = getstatefromvarargs({...})
	local limit = args[1]
	return internal_capped(s, limit)
end)

// P_RandomRange()
rawset(_G, "KR_RandomRange", function(...)
	local s, args = getstatefromvarargs({...})
	local upper = args[2]
	local lower = args[1]
	if lower > upper then print("invalid range for KR_RandomRange") return 0
	elseif lower == upper then return lower end
	if ((upper-lower) < 0) or ((upper-lower) == INT32_MAX) then
		print("lower and upper are too far apart")
		return 0
	end
	return internal_capped(s, (upper-lower)+1, 1) + lower
end)

// SG_Archive stuff, thanks GHG
if not SG_Archive then
	local rp

	-- SG_Archive must ONLY be called in MapLoad hooks. Players are not yet initialized in MapChange!
	-- Be careful if calling SG_Archive conditionally, the data may be saved/loaded in the wrong order!
	rawset(_G, "SG_Archive", function(var)
		if isdedicatedserver then return var end
		if replayplayback then
			if not (rp and rp.valid) then
				for p in players.iterate do
					if p.sgreplay then
						rp = p
						print("Loading replay data from "..rp.name)
						break
					end
				end
			end
			return table.remove(rp.sgreplay, 1)
		else
			if not (rp and rp.valid) then
				rp = consoleplayer
				rp.sgreplay = {}
				print("Saving replay data to "..rp.name)
			end
			table.insert(rp.sgreplay, var)
			return var
		end
	end)

	-- don't clog up replay data
	addHook("MapLoad", function()
		if not replayplayback and rp and rp.valid then rp.sgreplay = {} end
	end)

	addHook("PreThinkFrame", function()
		if replayplayback and rp and rp.valid and leveltime > 0 and next(rp.sgreplay) then
			print("\130Stale replay data found! Please report this issue!")
		end
	end)
end

// This is for saving the kartrand_global state for replays.
// You'll need to do this yourself for your own states
local function kartrand_global_archive()
	// archiving made easy thank you ghg
	kartrand_global = SG_Archive($)
end

addHook("MapLoad", kartrand_global_archive)

// For netvars, you'll also need to do this yourself
local function kartrand_global_sync(net)
	kartrand_global = net($)
end

addHook("NetVars", kartrand_global_sync)

// Now that we're done...
print("Testing random functions...")
print("KR_Random(): "..KR_Random())
print("KR_RandomFixed(): "..KR_RandomFixed())
print("KR_RandomByte(): "..KR_RandomByte())
print("KR_RandomSignedByte(): "..KR_RandomSignedByte())
print("KR_RandomBool(): "..tostring(KR_RandomBool()))
print("KR_RandomKey(604): "..KR_RandomKey(604))
print("KR_RandomRange(69, 420): "..KR_RandomRange(69, 420))

// Passed the test, lets prevent this from loading again.
rawset(_G, "kartrand", 1)